@echo off
setlocal
REM
REM Script : esilo_backup_db_mssql.cmd
REM Version: 2.0.3
REM Purpose: Use Microsoft SQL Server (osql.exe) to perform hot database backups
REM

REM ************************** BEGIN CONFIGURATION **************************

REM Space-separated list of databases to backup in the form:
REM SERVERNAME for the default instance, or
REM SERVERNAME\INSTANCE for a specific database instance
REM
set DATABASE_SERVERS_AND_INSTANCES=localhost

REM The location of this script and supporting programs
REM
set SCRIPT_DIR=D:\esilo_backup_db_mssql

REM The location to write out database backup files
REM This could be a local path (e.g. D:\database_backups),
REM or a UNC path (e.g. \\server\share\folder)
REM
set BACKUP_DIR=D:\esilo_backup_db_mssql

REM Files must be split into parts of 64 GiB or less. Common sizes:
REM 64 GiB: 68719476736 32 GiB: 34359738368 700 MiB: 734003200
REM Must be between 1073741824 and 68719476736; aborting.
REM
set SPLIT_SIZE=34359738368

REM Path to the osql.exe program
REM If you have more than one version of SQL Server installed, you can set
REM this to the full path of the osql.exe for the version you want to use.
REM
set OSQLEXE=osql.exe

REM The authentication method to use when connecting to the database. Options:
REM WINDOWS - for Windows authentication (no username or password required)
REM SQL     - for SQL Server authentication (requires username and password)
REM In most cases, Windows authentication will work. If you get access denied errors,
REM then you will probably need to use SQL authentication
REM
set SQL_AUTHMODE=WINDOWS
REM set SQL_AUTHMODE=SQL
REM set SQL_USERNAME=sa
REM set SQL_PASSWORD=password

REM If you need to access a network path, supply the network path with NET_PATH
REM If you need to map that network path to a drive letter, set the drive with NET_DRIVE
REM If you need to authenticate to access the network path, specify a username with NET_USERNAME
REM If you need a password to authenticate, specify a password with NET_PASSWORD
REM set NET_PATH=\\server\share\folder
REM set NET_DRIVE=Z
REM set NET_USERNAME=myuser
REM set NET_PASSWORD=MyPassword123


REM *************************** END CONFIGURATION ***************************








REM Set some variables
SET TEE=%SCRIPT_DIR%\wtee.exe
SET SPLITEXE=%SCRIPT_DIR%\esplit.exe
SET LOG=%BACKUP_DIR%\backup.log
SET SELF=%SCRIPT_DIR%\esilo_backup_db_mssql.cmd
SET SCRIPT_TITLE=eSilo MSSQL Database Backup Script

REM Create backup directory if it does not already exist
if not exist %BACKUP_DIR% mkdir %BACKUP_DIR%

REM Start log
echo:>%LOG%
echo %DATE% %TIME% ] %SCRIPT_TITLE% Started; Running sanity checks...
echo %DATE% %TIME% ] %SCRIPT_TITLE% Started; Running sanity checks... > %LOG%

REM Sanity checks
REM OSQL output: Microsoft (R) SQL Server Command Line Tool
%OSQLEXE% /? | findstr /B /C:"Microsoft (R) SQL Server Command Line Tool" >nul
if %ERRORLEVEL% EQU 0 goto osql_ok
call :fail OSQL program not found or unexpected version
exit /b 1
:osql_ok

if not exist %TEE% goto err_tee
if not exist %SPLITEXE% goto err_split
goto err_ok
:err_tee
echo %DATE% %TIME% ] Fatal Error: File does not exist: %TEE%
echo %DATE% %TIME% ] Fatal Error: File does not exist: %TEE% >> %LOG%
eventcreate /l Application /t ERROR /so "%SCRIPT_TITLE%" /id 1 /d "%SCRIPT_TITLE% Failed: Fatal Error: File does not exist: %TEE%. See log file %LOG% for more details." 2>&1 >nul
exit /b 1
:err_split
call :fail Fatal Error: File does not exist: %SPLITEXE%
exit /b 1
:err_ok

REM Check split size
set SPLIT_SIZE_MIN=1073741824
set SPLIT_SIZE_MAX=68719476736
call :isgreaterthan RESULT SPLIT_SIZE SPLIT_SIZE_MIN 
if %RESULT%. == 0. goto splitsize_min_ok
if %RESULT%. == 1. goto splitsize_min_ok
call :fail Split size too small: %SPLIT_SIZE% -- Must be between 1073741824 and 68719476736; aborting.
exit /b 1
:splitsize_min_ok
call :isgreaterthan RESULT SPLIT_SIZE SPLIT_SIZE_MAX
if %RESULT%. == 0. goto splitsize_max_ok
if %RESULT%. == -1. goto splitsize_max_ok
call :fail Split size too large: %SPLIT_SIZE% -- Must be between 1073741824 and 68719476736; aborting.
exit /b 1
:splitsize_max_ok

REM Set up osql authentication arguments
if %SQL_AUTHMODE%. == WINDOWS. goto auth_win
if %SQL_AUTHMODE%. == SQL. goto auth_sql
call :fail Invalid authentication method '%SQL_AUTHMODE%' ; aborting.
exit /b 1
:auth_win
set OSQL_AUTHOPTS=-E
set OSQL_AUTHOPTS_MASKED=-E
goto auth_done
:auth_sql
if %SQL_USERNAME%. == . goto auth_nouser
if %SQL_PASSWORD%. == . goto auth_nopass
set OSQL_AUTHOPTS=-U %SQL_USERNAME% -P %SQL_PASSWORD%
set OSQL_AUTHOPTS_MASKED=-U %SQL_USERNAME% -P PASSWORD
goto auth_done
:auth_nouser
call :fail SQL username is required when using SQL authentication, but none was provided; aborting.
exit /b 1
:auth_nopass
call :fail SQL password is required when using SQL authentication, but none was provided; aborting.
exit /b 1
:auth_done

REM Network drive mounting
if %NET_PATH%. == . goto net_done
if %NET_DRIVE%. == . goto net_drive_missing

set NETUSE_DRIVEOPTS= %NET_DRIVE%: 
goto net_drive_done

:net_drive_missing
set NETUSE_DRIVEOPTS=
:net_drive_done

if %NET_USERNAME%. == . goto net_user_missing
if %NET_PASSWORD%. == . goto net_pass_missing

set NETUSE_OPTS= /USER:%NET_USERNAME% %NET_PASSWORD% 
goto net_mount

:net_user_missing
set NETUSE_OPTS=
goto net_mount

:net_pass_missing
set NETUSE_OPTS= /USER:%NET_USERNAME% 
goto net_mount


:net_mount
echo %DATE% %TIME% ] Attempting to mount %NET_PATH% on %NET_DRIVE%: ... | %TEE% -a %LOG%
echo Y | net use %NETUSE_DRIVEOPTS% /DELETE 2>&1 >nul
echo Y | net use %NETUSE_DRIVEOPTS% %NET_PATH% %NETUSE_OPTS% > %LOG%.tmp
set EL=%ERRORLEVEL%
type %LOG%.tmp | %TEE% -a %LOG%
del /Q %LOG%.tmp
if %EL%. == 0. goto net_ok
call :fail Unable to mount network path (%NET_PATH%) to %NET_DRIVE%; aborting.
exit /b 1

goto net_done
:net_ok
echo %DATE% %TIME% ] Successfully mounted %NET_PATH% on %NET_DRIVE%: | %TEE% -a %LOG%
:net_done



echo %DATE% %TIME% ] %SCRIPT_TITLE% Sanity check completed successfully. | %TEE% -a %LOG%
echo: | %TEE% -a %LOG%
echo Configuration: | %TEE% -a %LOG%
echo: | %TEE% -a %LOG%
echo Script directory: %SCRIPT_DIR% | %TEE% -a %LOG%
echo Backup directory: %BACKUP_DIR% | %TEE% -a %LOG%
echo Tee program.....: %TEE% | %TEE% -a %LOG%
echo Split program...: %SPLITEXE% | %TEE% -a %LOG%
echo Split size......: %SPLIT_SIZE% | %TEE% -a %LOG%
echo Log file........: %LOG% | %TEE% -a %LOG%
echo OSQL path.......: %OSQLEXE% | %TEE% -a %LOG%
echo Auth method.....: %AUTH_METHOD% | %TEE% -a %LOG%
echo Username........: %SQL_USERNAME% | %TEE% -a %LOG%
if %SQL_PASSWORD%. == . goto pass_no
echo Password........: Yes | %TEE% -a %LOG%
goto pass_done
:pass_no
echo Password........: No | %TEE% -a %LOG%
:pass_done
echo Network Path....: %NET_PATH% | %TEE% -a %LOG%
if %NET_USERNAME%. == . goto netuser_no
echo Network Username: %NET_USERNAME% | %TEE% -a %LOG%
goto netuser_done
:netuser_no
echo Network Username: (none) | %TEE% -a %LOG%
:netuser_done
if %NET_PASSWORD%. == . goto netpass_no
echo Network Password: Yes | %TEE% -a %LOG%
goto netpass_done
:netpass_no
echo Network Password: No | %TEE% -a %LOG%
:netpass_done



echo: | %TEE% -a %LOG%




REM Main loop: for each instance...
set ENDRESULT=0
for %%d in (%DATABASE_SERVERS_AND_INSTANCES%) do ( call :doinstance %%d 
if ERRORLEVEL 1 SET ENDRESULT=1
)
if %ENDRESULT% EQU 0 goto ok
echo: | %TEE% -a %LOG%
call :fail ERROR: One or more databases were not backed up (code: %ENDRESULT%)
exit /b 1






REM ***************** BEGIN DOINSTANCE ***************** 
:doinstance
set SERVER_INSTANCE=%1

if not exist %BACKUP_DIR%\backups mkdir %BACKUP_DIR%\backups
if not exist %BACKUP_DIR%\backups\%SERVER_INSTANCE% mkdir %BACKUP_DIR%\backups\%SERVER_INSTANCE% 
set INSTANCE_BACKUP_DIR=%BACKUP_DIR%\backups\%SERVER_INSTANCE%
set OSQL_OPTS=%OSQL_AUTHOPTS% -S %SERVER_INSTANCE%
set OSQL_OPTS_MASKED=%OSQL_AUTHOPTS_MASKED% -S %SERVER_INSTANCE%


REM Delete old backups
if exist %INSTANCE_BACKUP_DIR%\*.bak  del /Q  %INSTANCE_BACKUP_DIR%\*.bak
if exist %INSTANCE_BACKUP_DIR%\*.bak.??? del /Q  %INSTANCE_BACKUP_DIR%\*.bak.???



echo: | %TEE% -a %LOG%
echo %DATE% %TIME% ] Backing up '%SERVER_INSTANCE%' with '%OSQLEXE% %OSQL_OPTS_MASKED%' ... | %TEE% -a %LOG%
echo: | %TEE% -a %LOG%

set BACKUP_BAK=1

(

echo DECLARE @DBName varchar(255 ^) 
echo DECLARE @DBFileName varchar(256 ^)
echo. 
echo DECLARE DATABASES_CURSOR CURSOR FOR
echo SELECT
echo     name
echo FROM
echo     sysdatabases
echo WHERE
echo     name
echo NOT IN ( 'tempdb' ^) 
echo.
echo OPEN DATABASES_CURSOR
echo.
echo FETCH NEXT FROM DATABASES_CURSOR INTO @DBName
echo.
echo WHILE @@FETCH_STATUS = 0
echo BEGIN
echo     set @DBFileName = replace(replace(@DBName,':','_' ^) ,'\','_' ^) + '.bak'
echo.
echo     exec ('BACKUP DATABASE [' + @DBName + '] TO  DISK = N''%BACKUP_DIR%\backups\%SERVER_INSTANCE%\' + 
echo         @DBFileName + ''' WITH NOFORMAT, INIT,  NAME = N''' + 
echo         @DBName + '-Full Database Backup'', SKIP, NOREWIND, NOUNLOAD,  STATS = 100' ^)
echo.
echo     FETCH NEXT FROM DATABASES_CURSOR INTO @DBName
echo END
echo.
echo CLOSE DATABASES_CURSOR
echo DEALLOCATE DATABASES_CURSOR
echo.

) >  %INSTANCE_BACKUP_DIR%\script.sql

type  %INSTANCE_BACKUP_DIR%\script.sql | %OSQLEXE% %OSQL_OPTS% 2>&1 > %LOG%.tmp
set EL=%ERRORLEVEL%
if %EL% NEQ 0 goto errorbackup
echo: | %TEE% -a %LOG%
echo: | %TEE% -a %LOG%
echo %DATE% %TIME% ] Backup of '%SERVER_INSTANCE%' completed successfully. Output: | %TEE% -a %LOG%
echo: | %TEE% -a %LOG%
type %LOG%.tmp | %TEE% -a %LOG%
echo: | %TEE% -a %LOG%
del /Q %LOG%.tmp



REM Split files
echo. | %TEE% -a %LOG%
echo %DATE% %TIME% ] Splitting database dump files for '%SERVER_INSTANCE%' ... | %TEE% -a %LOG%
echo. | %TEE% -a %LOG%
for %%i in (%INSTANCE_BACKUP_DIR%\*.bak) do (
call :splitfile %%i
)

echo. | %TEE% -a %LOG%
echo %DATE% %TIME% ] Split of files for '%SERVER_INSTANCE%' Completed successfully. | %TEE% -a %LOG%
del /Q  %INSTANCE_BACKUP_DIR%\script.sql


)


goto :eof
REM ***************** END DOINSTANCE ***************** 





:errorbackup
call :fail Failed backing up %SERVER_INSTANCE% (exit code %EL%).
exit /b 1

:ok


echo: | %TEE% -a %LOG%
echo %DATE% %TIME% ] %SCRIPT_TITLE% Completed successfully. | %TEE% -a %LOG%
eventcreate /l Application /t INFORMATION /so "%SCRIPT_TITLE%" /id 1 /d "%SCRIPT_TITLE% Completed Successfully." 2>&1 >nul
exit /b 0





REM ************** FUNCTIONS ************** 


:splitfile
	set FILESIZE=%~z1%
	echo %DATE% %TIME% ] File size of %1 is %FILESIZE% | %TEE% -a %LOG%
	call :isgreaterthan RESULT FILESIZE SPLIT_SIZE
	if %RESULT%. == 1. goto filesize_gtr
	echo %DATE% %TIME% ] File size is less than or equal to max file size; Renaming instead of splitting... | %TEE% -a %LOG%
	move %1 %1.000 2>&1 >nul
	if not "%ERRORLEVEL%" == "0" goto rename_error
	echo %DATE% %TIME% ] Successfully renamed file. | %TEE% -a %LOG%
	goto filesize_done
	:rename_error
	echo %DATE% %TIME% ] ERROR: Rename of file %1 to %1.000 failed. | %TEE% -a %LOG%
	eventcreate /l Application /t ERROR /so "%SCRIPT_TITLE%" /id 1 /d "%SCRIPT_TITLE% Failed: Rename of file %%i to %%i.000 failed. See log file %LOG% for more details." 2>&1 >nul
	exit /b 1
	
	:filesize_gtr
	echo %DATE% %TIME% ] File size is greater than %SPLIT_SIZE% ; Splitting %1 ... | %TEE% -a %LOG%
	%SPLITEXE% %1 %SPLIT_SIZE%   2>&1 | %TEE% -a %LOG%
	IF NOT "%ERRORLEVEL%" == "0" goto spliterror
	IF NOT EXIST %1.000 goto spliterror
	del /Q %1 | %TEE% -a %LOG%
	:filesize_done
	exit /b 0
	
	:spliterror
	
	echo %DATE% %TIME% ] ERROR: Split of file(s) failed. | %TEE% -a %LOG%
	eventcreate /l Application /t ERROR /so "%SCRIPT_TITLE%" /id 1 /d "%SCRIPT_TITLE% Failed: Splitting of file(s) failed. See log file %LOG% for more details." 2>&1 >nul
	exit /b 1



:fail <stringVar>
	echo %DATE% %TIME% ] %* | %TEE% -a %LOG%
	eventcreate /l Application /t ERROR /so "%SCRIPT_TITLE%" /id 1 /d "%SCRIPT_TITLE% Failed: %*. See log file %LOG% for more details." 2>&1 >nul
(
	endlocal
	exit /b 1
)


:strlen <resultVar> <stringVar>
(   
    setlocal EnableDelayedExpansion
    set "s=!%~2!#"
    set "len=0"
    for %%P in (4096 2048 1024 512 256 128 64 32 16 8 4 2 1) do (
        if "!s:~%%P,1!" NEQ "" ( 
            set /a "len+=%%P"
            set "s=!s:~%%P!"
        )
    )
)
( 
	endlocal
	set "%~1=%len%"
	exit /b
)




:isgreaterthan <resultVar> <stringVar> <stringVar>
	setlocal EnableDelayedExpansion
	SET "L=!%~2!"
	SET "R=!%~3!"
	
	call :strlen LENGTH L

	call :strlen MAXLENGTH R

	if %LENGTH% GTR %MAXLENGTH% goto greater
	if %LENGTH% LSS %MAXLENGTH% goto less

	REM Do a string comparison. Append a non-digit character and compare again
	if %L%a GTR %R%a goto greater
	if %L%a LSS %R%a goto less
	set RESULT=0
	goto fdone

	:less
	set RESULT=-1
	goto fdone

	:greater
	set RESULT=1
	:fdone

(
	endlocal
	set "%~1=%RESULT%"
	exit /b
)





